/*
  Fast hash table implementation
  Max Volkov (Floyd_Rose), Sevastopol - 2007
  Written for XFLib project
  Mailto: max at xfhome dot org
*/

#include <stdlib.h>

#include "hash.h"

int hash_alloc_counter = 0;
int hash_alloc_sum = 0;
int hash_free_counter = 0;

hash_user_alloc_t hash_user_alloc = HASH_NULL;
hash_user_free_t  hash_user_free  = HASH_NULL;

/*
  Allocate memory
*/
void* hash_alloc(int size)
{
 hash_alloc_counter++;
 hash_alloc_sum += size;

 if (!hash_user_alloc) return malloc(size);

 return hash_user_alloc(size);
}

/*
  Free memory
*/
void hash_free(void *p)
{
 hash_free_counter++;

 if (!hash_user_free) free(p);
 else hash_user_free(p);
}

/*
  Create new hash table
*/
hash_t* hash_create(int hash_size)
{
 int i;
 hash_t *hash;
 void **p;

 if (hash_size <= 0) return HASH_NULL;

 hash = (hash_t*) hash_alloc(sizeof(hash_t));
 if (!hash) return HASH_NULL;

 hash->size = 0;

 hash->block = (void**) hash_alloc(sizeof(void*) * hash_size);
 if (!hash->block) { hash_free(hash); return 0; }

 for (i = 0; i < hash_size; i++)
  {
   p = hash->block + i; *p = 0;
  }

 hash->size = hash_size;
 hash->magic = HASH_MAGIC;

 return hash;
}

/*
  Destroy hash table and free all data
*/
void hash_destroy(hash_t *hash)
{
 int i,j,size;
 hash_block *bk,*_bk;

 if (!hash) return;
 if (!hash->size) return;

 /* Free all block queues in hash table */
 for (i = 0; i < hash->size; i++)
  {
   bk = hash->block[i];
   _bk = bk;

   while (_bk)
    {
     bk = _bk;
     _bk = (hash_block*)bk->next;

     /* Free key */
     if (bk->attr & HASH_COPY_KEY)
       if (bk->key)
         hash_free(bk->key);

     /* Free data */
     if (bk->attr & HASH_COPY_DATA)
       if (bk->data)
         hash_free(bk->data);

     hash_free(bk);
    }
  }

 /* Free hash strcuture */
 hash_free(hash->block);
 hash_free(hash);
}

/*
  Duplicate memory area
*/
void* hash_memdup(void *data,int size)
{
 void *p;
 int s;

 s = size * sizeof(void*);

 p = hash_alloc(s);
 if (!p) return HASH_NULL;
 memcpy(p,data,s);

 return p;
}

/*
  Add element to the hash table
*/
hash_t* hash_add(hash_t *hash,void *key,int keysize,void *data,int datasize,int attr)
{
 int bkn,sum,al = 0,f = 0, q = 0;
 hash_block *bk,*_bk,*lastbk = 0;

 if (!hash) return;

 sum = hash_sum(key,keysize);
 bkn = hash_function(hash->size,key,keysize);

 if (!hash->block[bkn])
  {
   /* Empty hash queue, add one element */
   bk = hash_alloc(sizeof(hash_block));
   if (!bk) return HASH_NULL;
   hash->block[bkn] = bk;
   bk->next = HASH_NULL;
   al = 1;
   q = 0;
  }
 else
  {
   /* Seek for the key in the existing queue */
   bk = (hash_block*)hash->block[bkn];
   _bk = bk;
   q = 1;

   while (_bk)
    {
     bk = _bk;
     if (bk->keyhash == sum)
       if (bk->keysize == keysize)
         if (!memcmp(bk->key,key,keysize))
          {
           /* Key found, exit if we do not need to replace it */
           if (!(attr & HASH_REPLACE_KEY)) return HASH_NULL;

           if (bk->key && bk->keysize && (bk->attr & HASH_COPY_KEY))
             hash_free(bk->key);

           if (bk->data && bk->datasize && (bk->attr & HASH_COPY_DATA))
             hash_free(bk->key);

           f = 1;
          }

     _bk = (hash_block*)bk->next;
    }

   lastbk = bk;

   if (!f)
    {
     /* Key not found, add one element */
     bk = hash_alloc(sizeof(hash_block));
     if (!bk) return HASH_NULL;
     bk->next = HASH_NULL;
     al = 1;
    }
  }

 /* Key not found, so add it;) */

 /* Do we need to copy key? */
 if (attr & HASH_COPY_KEY)
  {
   if (key && keysize)
    {
     key = hash_memdup(key,keysize);
     if (!key)
      {
       if (al) { hash->block[bkn] = HASH_NULL; hash_free(bk); }
       return HASH_NULL;
      }
    }
   else { attr = attr & (~HASH_COPY_KEY); key = HASH_NULL; }
  }
 else
  {
   if (!key || !keysize) { key = HASH_NULL; keysize = HASH_NULL; }
  }

 /* Do we need to copy data? */
 if (attr & HASH_COPY_DATA)
  {
   if (data && datasize)
    {
     data = hash_memdup(data,datasize);
     if (!data)
      {
       if (al) { hash->block[bkn] = HASH_NULL; hash_free(bk); }
       if (attr & HASH_COPY_KEY) hash_free(key);
       return HASH_NULL;
      }
    }
   else { attr = attr & (~HASH_COPY_DATA); data = HASH_NULL; }
  }
 else
  {
   if (!data || !datasize) { data = HASH_NULL; datasize = HASH_NULL; }
  }

 /* Fill element data */
 bk->magic = HASH_MAGIC;
 bk->keyhash = sum;
 bk->key = key;
 bk->keysize = keysize;
 bk->data = data;
 bk->datasize = datasize;
 bk->attr = attr;

 /* Set last block pointer */
 if (al)
  {
   if (!q) hash->block[bkn] = bk;
   else lastbk->next = bk;
  }

 return hash;
}

/*
  Hash function
*/
int hash_function(int hashsize,void *string,int size)
{
 return abs(hash_sum(string,size) % hashsize);
}

/*
  String hash sum function
*/
int hash_sum(void *string,int size)
{
 char *s = string;
 signed int i,sum;

 sum = 0x25051987;

 for (i = 0; i < size; i++)
  {
   sum *= 31;
   sum += s[i];
  }

 return sum;
}

/*
  Fetch data
*/
void* hash_find(hash_t *hash,void *key,int keysize,int *datasize)
{
 int bkn,sum;
 hash_block *bk,*_bk;

 sum = hash_sum(key,keysize);
 bkn = hash_function(hash->size,key,keysize);

 /* Seek for the key in the existing queue */
 bk = (hash_block*)hash->block[bkn];
 if (!bk) return HASH_NULL;

 _bk = bk;

 while (_bk)
  {
   bk = _bk;

   if (bk->keyhash == sum)
     if (bk->keysize == keysize)
       if (!memcmp(bk->key,key,keysize))
         {
          /* Key found, exit if we do not need to return it */
          if (datasize) *datasize = bk->datasize;
          return bk->data;
         }

   _bk = (hash_block*)bk->next;
  }

 return HASH_NULL;
}

/*
  Remove hash table element
*/
hash_t *hash_remove(hash_t *hash,void *key,int keysize)
{
 int bkn,sum,fst = 1;
 hash_block *bk,*_bk,*lastbk;

 sum = hash_sum(key,keysize);
 bkn = hash_function(hash->size,key,keysize);

 /* Seek for the key in the existing queue */
 bk = (hash_block*)hash->block[bkn];
 if (!bk) return HASH_NULL;

 _bk = bk;

 while (_bk)
  {
   bk = _bk;

   if (bk->keyhash == sum)
     if (bk->keysize == keysize)
       if (!memcmp(bk->key,key,keysize))
         {
          /* Key found, exit if we do not need to remove block */
          if (bk->key && bk->keysize && (bk->attr & HASH_COPY_KEY))
            hash_free(bk->key);

          if (bk->data && bk->datasize && (bk->attr & HASH_COPY_DATA))
            hash_free(bk->data);

          /* If we remove first block */
          if (fst)
           {
            hash->block[bkn] = bk->next;
           }
          else
           {
            lastbk->next = bk->next;
           }

          hash_free(bk);

          return hash;
         }

   _bk = (hash_block*)bk->next;
   fst = 0;
   lastbk = bk;
  }

 return HASH_NULL;
}

#ifdef HASH_DEBUG

/*
  Print hash dump (debugging)
*/
void hash_print(hash_t *hash)
{
 int i;
 hash_block *bk,*_bk;

 for (i = 0; i < hash->size; i++)
  {
   printf("Block %i\n",i);

   bk = (hash_block*)hash->block[i];
   _bk = bk;

   while (_bk)
    {
     bk = _bk;
     printf("\t%s=%s\n",bk->key,bk->data);
     _bk = (hash_block*)bk->next;
    }
  }
}

void hash_info()
{
 printf("hash_alloc_counter = %i\n",hash_alloc_counter);
 printf("hash_free_counter  = %i\n",hash_free_counter);
 printf("hash_alloc_sum     = %i\n",hash_alloc_sum);
}

#endif

/*
  Duplicate hash table
*/
hash_t* hash_dup(hash_t *hash,int hash_size)
{
 hash_t *h;
 hash_block *bk;
 int i;

 h = hash_create(hash_size);
 if (!h) return HASH_NULL;

 for (i = 0; i < hash->size; i++)
  {
   bk = (hash_block*)hash->block[i];
   while (bk)
    {
     if (!hash_add(h,bk->key,bk->keysize,bk->data,bk->datasize,bk->attr))
      {
       hash_destroy(h);
       return HASH_NULL;
      }
     bk = (hash_block*)bk->next;
    }
  }

 return h;
}

/*
  Resize hash table
*/
hash_t *hash_resize(hash_t *hash,int hash_size)
{
 hash_t *h;
 void **p;
 int t;

 h = hash_dup(hash,hash_size);
 if (!h) return HASH_NULL;

 t = hash->size;  hash->size  = h->size;  h->size  = t;
 p = hash->block; hash->block = h->block; h->block = p;

 hash_destroy(h);

 return hash;
}


