/*
  XFLib - XF management library

  (C) Max Volkov, Sevastopol 2006-2008
  Mailto: max at xfhome dot org
*/

/*
  Core XF data model engine
*/

#include <string.h>
#include <stdlib.h>

#include "xf.h"
#include "xf_api.h"
#include "xf_sys.h"

#include "hash.h"

/*
  Allocate memory
*/
_XF_API xfVoid*
xfAlloc(xfInt size)
{
 xfAllocCounter++;
 xfAllocSum += size;
 if (!xfUserAlloc) return malloc(size);
 return xfUserAlloc(size);
}

/*
  Free memory
*/
_XF_API xfVoid
xfFree(xfVoid *p)
{
 xfFreeCounter++;
 if (!xfUserFree) free(p);
 else xfUserFree(p);
}

/*
  Allocates new node
*/

_XF_SYS xfNode*
xfAllocNode()
{
 return (xfNode*)xfAlloc(sizeof(xfNode));
}

/*
  Adds node
*/

_XF_API xfNode*
xfAdd(xfNode *parent,xfChar *name,xfChar *value,xfChar *Class)
{
 xfNode *node,*c;
 xfBool found;
 xfSecurity *sec;
 xfHash *hash;

 if (!parent) return XF_NULL;
 if (!name) return XF_NULL;

 /* Check if node already exists */

 if (!parent->hash)
  {
   c = parent->firstChild;
   found = false;
   while (c)
    {
     if (c->name)
       if (!xfStrcmp(c->name,name))
         { found = true; break; }
     c = c->next;
    }
  }
 else
  {
   found = false;
   c = xfHashFindNode(parent->hash,name);
   if (c) found = true;
  }

 if (found) return c;

 node = xfAllocNode();
 if (!node) return XF_NULL;

 name = xfStrdup(name);
 if (!name)
  {
   xfFree(node);
   return XF_NULL;
  }

 node->Class = XF_NULL;
 if (Class)
   if (*Class)
    {
     Class = xfStrdup(Class);
     if (!Class)
      {
       xfFree(node);
       xfFree(Class);
       return XF_NULL;
      }
     node->Class = Class;
    }

 node->value = XF_NULL;
 if (value)
   if (*value)
    {
     value = xfStrdup(value);
     if (!value)
      {
       xfFree(node);
       xfFree(name);
       if (Class) xfFree(Class);
       return XF_NULL;
      }
     node->value = value;
    }

 if (found) return node;

 node->map = parent->map;
 node->parent = parent;
 node->name = name;

 /* Set node information */
 node->prev = parent->lastChild;
 node->next = XF_NULL;
 node->firstChild = XF_NULL;
 node->lastChild = XF_NULL;
 node->numChildren = 0;

 node->hash = XF_NULL;

 /* If tree has nodes, set the last one to point to the new one */
 if (parent->firstChild) parent->lastChild->next = node;
 else parent->firstChild = parent->lastChild = node;

 /* Increment parent's number of children */
 parent->lastChild = node;
 parent->numChildren++;

 /*
   If hash table of parent node exists, add new node
 */
 if (parent->hash)
  {
   xfHashAddNode(parent->hash,node);

   /* If we have XF_AUTOHASH option, we need to resize hash
      automatically each time number of elements in the hash
      overreaches hash table size */
   #ifdef XF_AUTOHASH
   if (parent->numChildren > ((xfHash*)parent->hash)->size)
    {
     hash = xfResizeHash((xfHash*)parent->hash,
                         parent->numChildren * XF_AUTOHASH);
     if (hash) parent->hash = hash;
    }
   #endif
  }
 else
  {
   /*
     If we have too many child nodes, we need
     to create hash table for child names
   */

   if (parent->numChildren >= XF_HASH_NODE_SIZE && XF_HASH_NODE_SIZE != 0)
    {
     xfCreateNodeHash(parent,XF_HASH_SIZE);
    }
  }

 ((xfMap*)parent->map)->nodeNum++;

 return node;
}

/*
  Remove node and all its content
*/
_XF_API xfVoid
xfRemove(xfNode *node)
{
 xfNode *c,*p;
 xfMap *m;

 if (!node) return;

 /* Remove node children (recurse) */
 c = node->firstChild;

 while (c)
  {
   p = c->next;
   xfRemove(c);
   c = p;
  }

 /* Free strings */
 if (node->name) xfFree(node->name);
 if (node->Class) xfFree(node->Class);
 if (node->value) xfFree(node->value);

 p = node->parent;
 if (p)
  {
   if (p->hash)
    {
     xfHashRemoveNode((xfHash*)p->hash,node);
    }
  }

 /* Free node children hash (if exists) */
 if (node->hash) xfDestroyHash((xfHash*)node->hash);

 /* Set pointer in the previous node */
 p = node->prev;
 if (p) p->next = node->next;

 p = node->next;
 if (p) p->prev = node->prev;

 m = (xfMap*)node->map;
 if (m->root != node)
  {
   /* Do we remove first node of some parent? */
   p = node->parent;
   if (p)
    {
     /* Do we need to update parent's firstChild pointer? */
     if (p->firstChild == node) p->firstChild = node->next;
     /* Do we need to update parent's lastChild pointer? */
     if (p->lastChild == node)  p->lastChild  = node->prev;
     /* Descement parent's number of children */
     p->numChildren--;
    }
  }

 ((xfMap*)node->map)->nodeNum--;

 /* Free node structure */
 xfFree(node);
}

/*
  Get first node child
*/
_XF_API xfNode*
xfParent(xfNode *node)
{
 xfMap *m;

 if (!node) return XF_NULL;

 m = (xfMap*)node->map;
 if (node->parent == m->root) return XF_NULL;

 return node->parent;
}

/*
  Get node name
*/
_XF_API xfChar*
xfName(xfNode *node)
{
 if (!node) return XF_NULL;
 return node->name;
}

/*
  Get first node child
*/
_XF_API xfNode*
xfChild(xfNode *node)
{
 xfMap *m;

 if (!node) return XF_NULL;

 return node->firstChild;
}

/*
  Get last node child
*/
_XF_API xfNode*
xfLastChild(xfNode *node)
{
 xfMap *m;

 if (!node) return XF_NULL;

 return node->lastChild;
}

/*
  Get next node child
*/
_XF_API xfNode*
xfNext(xfNode *node)
{
 xfMap *m;

 if (!node) return XF_NULL;

 return node->next;
}

/*
  Get previous node child
*/
_XF_API xfNode*
xfPrev(xfNode *node)
{
 xfMap *m;

 if (!node) return XF_NULL;

 return node->prev;
}

/*
  Get node value
*/

_XF_API xfChar*
xfGet(xfNode *node)
{
 if (node) return node->value;
 return XF_NULL;
}

/*
  Set node value
*/

_XF_API xfChar*
xfSet(xfNode *node,xfChar *value)
{
 xfChar *s;

 if (!node) return XF_NULL;

 if (value)
  {
   s = xfStrdup(value);
   if (!s) return XF_NULL;
  }

 if (node->value) xfFree(node->value);

 node->value = XF_NULL;
 if (value) node->value = s;

 return node->value;
}

/*
 Get node class
*/

_XF_API xfChar*
xfGetClass(xfNode *node)
{
 if (node) return node->Class;
 return XF_NULL;
}

/*
  Set node class
*/

_XF_API xfChar*
xfSetClass(xfNode *node,xfChar *Class)
{
 xfChar *s;

 if (!node) return XF_NULL;

 if (Class)
  {
   s = xfStrdup(Class);
   if (!s) return XF_NULL;
  }

 if (node->Class) xfFree(node->Class);

 if (Class) node->Class = s;

 return node->Class;
}

/*
  Checks if node is global (root)
*/
_XF_API xfInt
xfIsRoot(xfNode *node)
{
 xfMap *m;

 if (!node) return false;

 m = (xfMap*)node->map;
 if (node->parent == m->root) return true;

 return false;
}

/*
  Gets node pointer by parent and name
*/
_XF_API xfNode*
xfEnter(xfNode *parent,xfChar *name)
{
 xfNode *node,*c;
 xfNode **cc;
 xfBool found;

 if (!parent) return XF_NULL;
 if (!name) return XF_NULL;
 if (!*name) return XF_NULL;

 /* If parent node has hash table, we try to find the name in it */
 if (parent->hash)
  {
   cc = (xfNode*)xfHashFindNode((xfHash*)parent->hash,name);
   return cc;
  }

 /* Find if node already exists */
 c = parent->firstChild;
 found = false;
 while (c)
  {
   if (c->name)
     if (!xfStrcmp(c->name,name))
       { found = true; break; }
   c = c->next;
  }

 if (!found) return XF_NULL;

 return c;
}

/*
  Returns number of node children
*/
_XF_API xfInt
xfNumChildren(xfNode *node)
{
 return node->numChildren;
}

/*
  Returns last document parse error as text message
*/
_XF_API xfChar*
xfGetError(xfMap *map)
{
 return map->error;
}

/*
  Sets parse error message
*/
_XF_SYS xfChar*
xfSetError(xfMap *map,xfChar *error)
{
 xfChar *msg;

 if (map->error) xfFree(map->error);
 map->error = XF_NULL;

 if (error)
  {
   msg = xfStrdup(error);
   map->error = msg;
  }
 else map->error = XF_NULL; 

 return msg;
}

/*
  Get error line number
*/
_XF_API xfInt
xfGetErrorLine(xfMap *map)
{
 return map->lineno;
}

/*
  Set error line number
*/
_XF_API xfInt
xfSetErrorLine(xfMap *map,xfInt lineno)
{
 map->lineno = lineno;
 return lineno;
}

/*
  Get error column number
*/
_XF_API xfInt
xfGetErrorCol(xfMap *map)
{
 return map->colno;
}

/*
  Set error column number
*/
_XF_API xfInt
xfSetErrorCol(xfMap *map,xfInt colno)
{
 map->colno = colno;
 return colno;
}


