/*
  XFLib - XF management library

  (C) Max Volkov, Sevastopol 2006-2008
  Mailto: max at xfhome dot org
*/

/*
  Input/output abstract layer routines
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "xf.h"
#include "xf_api.h"
#include "xf_sys.h"

_XF_SYS xfReader*
xfCreateReader(xfInt type,xfVoid *source)
{
 xfReader *r;
 char c[3] = {0x00, 0x00, 0x00};
 xfInt n;

 r = (xfReader*)xfAlloc(sizeof(xfReader));
 if (!r) return XF_NULL;

 if (type != XF_READER_MEMORY &&
     type != XF_READER_FILE)
  {
   /* Incorrect reader type */
   return XF_NULL;
  }

 r->bufSize = 0;
 r->pos = 0;
 r->string = 0;
 r->type = type;
 r->bufSize = 0;
 r->bufPos = 0;
 r->byteSwap = 0;

 if (type == XF_READER_MEMORY)
  {
   r->string = (xfChar*)source;
  }

 if (type == XF_READER_FILE)
  {
   r->file = fopen((xfNChar*)source,"rb");
   if (!r->file)
    {
     xfFree(r);
     return XF_NULL;
    }

   fread(c,1,3,r->file);

   r->fileType = XF_READER_FILE_ASCII;

   /*
     Detect file encoding by forst file bytes:
     0xFF 0xFE:      UTF-16 little endian
     0xFE 0xFF:      UTF-16 big endian
     0xEF 0xBB 0xBF: UTF-8
     OTHER:          ASCII
   */
   if ((char)c[0] == (char)0xFF && (char)c[1] == (char)0xFE)
     r->fileType = XF_READER_FILE_UTF16LE;
   if ((char)c[0] == (char)0xFE && (char)c[1] == (char)0xFF)
     r->fileType = XF_READER_FILE_UTF16BE;
   if ((char)c[0] == (char)0xEF &&
       (char)c[1] == (char)0xBB &&
       (char)c[2] == (char)0xBF)
     r->fileType = XF_READER_FILE_UTF8;

   if (r->fileType == XF_READER_FILE_UTF16BE)
      if (xfLittleEndian()) r->byteSwap = 1;
   if (r->fileType == XF_READER_FILE_UTF16LE)
      if (xfBigEndian()) r->byteSwap = 1;

   if (r->fileType != XF_READER_FILE_UTF8)
    {
     /* Reopen file as text from the beginning for ASCII files */
     fclose(r->file);
     r->file = fopen((xfChar*)source,"rb");
     if (!r->file)
      {
       xfFree(r);
       return XF_NULL;
      }
     if (r->fileType == XF_READER_FILE_UTF16BE ||
         r->fileType == XF_READER_FILE_UTF16LE)
       fread(c,2,1,r->file);
    }
  }

 return r;
}

_XF_SYS xfVoid
xfDestroyReader(xfReader *reader)
{
 if (reader->type == XF_READER_FILE)
   if (reader->file)
     fclose(reader->file);

 xfFree(reader);
}

_XF_SYS xfInt
xfReaderGetSize(xfReader *reader)
{
 int size,pos;

 if (!reader) return XF_NULL;

 if (reader->type == XF_READER_MEMORY)
   return xfStrlen(reader->string);

 if (reader->type == XF_READER_FILE) {
   pos = ftell(reader->file);
   fseek(reader->file,0,SEEK_END);
   size = ftell(reader->file);
   fseek(reader->file,pos,SEEK_SET);
   return size;
  }

 return XF_NULL;
}

_XF_SYS xfChar
xfReaderInputChar(xfReader *reader)
{
 xfInt c = 0x0000;
 xfChar wc,wcc[2];
 xfNChar cc[5] = {'\x00', '\x00', '\x00', '\x00', '\x00'};

 if (reader->type == XF_READER_MEMORY)
  {
   c = reader->string[reader->pos];
   if (c) reader->pos++;
   return c;
  }

 if (reader->type == XF_READER_FILE)
  {
   if (reader->fileType == XF_READER_FILE_ASCII)
    {
     c = fgetc(reader->file);
     if (c == EOF) return (xfChar)0x0000;
     if (!(c & 0x80)) return c;
     else return (xfChar)'?';
    }

   if (reader->fileType == XF_READER_FILE_UTF8)
    {
     cc[0] = fgetc(reader->file);
     if (cc[0] == EOF) return (xfChar)0x0000;

     if (!(cc[0] & 0x80)) return (xfChar)cc[0];

     if ((xfChar)(cc[0] & 0xE0) == (xfChar)(0xC0))
      {
       cc[1] = fgetc(reader->file);
       if (cc[1] == EOF) return (xfChar)0x0000;
       xfConvertFromUTF8(cc,&wcc);
       return wcc[0];
      }

     if ((xfChar)(cc[0] & 0xF0) == (xfChar)(0xE0))
      {
       cc[1] = fgetc(reader->file);
       if (cc[1] == EOF) return (xfChar)0x0000;
       cc[2] = fgetc(reader->file);
       if (cc[2] == EOF) return (xfChar)0x0000;
       xfConvertFromUTF8(cc,&wcc);
       return wcc[0];
      }

     return (xfChar)0x0000;
    }

   if (reader->fileType == XF_READER_FILE_UTF16BE ||
       reader->fileType == XF_READER_FILE_UTF16LE)
    {
     if (feof(reader->file)) return XF_NULL;
     if (fread(&wc,2,1,reader->file) != 1) return XF_NULL;

     if (reader->byteSwap) wc = xfCharByteSwap(wc);
     return wc;
    }
  }

 return c;
}

_XF_SYS xfWriter*
xfCreateWriter(xfInt type,xfVoid *dest,xfInt maxSize,xfInt charCodes)
{
 xfWriter *w;
 xfChar c[3] = {0x00, 0x00, 0x00};
 xfInt n;

 if (type != XF_WRITER_MEMORY &&
     type != XF_WRITER_FILE_ASCII &&
     type != XF_WRITER_FILE_UTF8 &&
     type != XF_WRITER_FILE_UTF16)
  {
   /* Incorretc writer type */
   return XF_NULL;
  }

 w = (xfWriter*)xfAlloc(sizeof(xfWriter));
 if (!w) return XF_NULL;

 w->string = 0;

 w->size = 0;
 w->maxSize = maxSize;
 w->type = type;
 w->shift = 0;

 w->o_charCodes = charCodes;

 if (w->type == XF_WRITER_MEMORY)
  {
   w->string = (xfChar*)dest;
   return w;
  }

 /*
   Open ASCII and UTF-8 files as normal text files
 */
 if (w->type == XF_WRITER_FILE_ASCII || w->type == XF_WRITER_FILE_UTF8)
  {
   w->file = fopen((xfChar*)dest,"wt");
   if (!w->file)
    {
     xfFree(w);
     return XF_NULL;
    }

   /* Write UTF-8 encoding header */
   if (w->type == XF_WRITER_FILE_UTF8)
     fwrite("\xEF\xBB\xBF",3,1,w->file);

   return w;
  }

 /*
   Open UTF-16 files as binary files
 */
 if (w->type == XF_WRITER_FILE_UTF16)
  {
   w->file = fopen((xfChar*)dest,"wb");
   if (!w->file)
    {
     xfFree(w);
     return XF_NULL;
    }

   /* Write UTF-16 encoding header (big or little endian) */
   if (xfBigEndian()) fwrite("\xFE\xFF",2,1,w->file);
   else fwrite("\xFF\xFE",2,1,w->file);
  }

 return w;
}

_XF_SYS xfVoid
xfDestroyWriter(xfWriter *writer)
{
 if (writer->type == XF_WRITER_FILE_ASCII ||
     writer->type == XF_WRITER_FILE_UTF8  ||
     writer->type == XF_WRITER_FILE_UTF16)
   if (writer->file)
     fclose(writer->file);

 xfFree(writer);
}

_XF_SYS xfVoid
xfWriterOutputString(xfWriter *writer,xfChar *string)
{
 xfInt i;
 xfNChar *b;

 if (!writer) return;

 if (writer->type == XF_WRITER_MEMORY)
  {
   for (i = 0; string[i] && writer->size + 1 < writer->maxSize; i++, writer->size++)
     writer->string[writer->size] = string[i];
   writer->string[writer->size] = L'\x0000';
   return;
  }

 if (writer->type == XF_WRITER_FILE_UTF16)
  {
   fwrite(string,xfStrlen(string) * sizeof(xfChar),1,writer->file);
  }

 if (writer->type == XF_WRITER_FILE_ASCII)
  {
   for (i = 0; string[i]; i++)
    {
     /* In ASCII we write only codes up to 127
        (greater code causes error, marked as '?') */
     if (!(string[i] & 0xFF80)) fputc(string[i],writer->file);
     else fputc('?',writer->file);
    }
  }

 if (writer->type == XF_WRITER_FILE_UTF8)
  {
   b = (xfNChar*)xfAlloc(xfStrlen(string) * 4 + 1);
   if (!b) return;
   xfConvertToUTF8(string,b);
   for (i = 0; b[i]; i++) fputc(b[i],writer->file);
   xfFree(b);
  }
}

