/*
  XFLib - XF management library

  (C) Max Volkov, Sevastopol 2006-2008
  Mailto: max at xfhome dot org
*/

/*
  XF document parser
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "xf.h"
#include "xf_api.h"
#include "xf_sys.h"

_XF_API xfInt
xfReadString(xfMap *map, xfChar *string)
{
 xfReader *r;
 xfInt res;

 if (!map) return XFE_DATA;

 r = xfCreateReader(XF_READER_MEMORY,(xfChar*)string);
 if (!r) return XFE_IO;

 res = xfSysParse(map,r);

 xfDestroyReader(r);

 return res;
}

_XF_API xfInt
xfReadFile(xfMap *map, xfNChar *filename)
{
 xfReader *r;
 xfInt res;

 if (!map) return XFE_DATA;

 r = xfCreateReader(XF_READER_FILE,(xfVoid*)filename);
 if (!r) return XFE_IO;

 res = xfSysParse(map,r);

 xfDestroyReader(r);

 return res;
}

/*
  XF document parser
  Supports two models:
   - event-based parsing (calls event handlers);
   - parsing to memory.
  Cancombine two parsing models
*/
_XF_SYS xfInt
xfSysParse(xfMap *map, xfReader *reader)
{
 #define ERROR(msg) \
  { \
    if (map) { \
    xfSetError(map,xfTranslateMsg(XFS(msg))); \
    xfSetErrorLine(map,line); \
    xfSetErrorCol(map,col); } \
    xfFree(word); \
    xfFree(stack); \
    xfFree(lstack); \
    return XFE_FORMAT; \
   }

 #define NO_COMMENT   0
 #define WIDE_COMMENT 1         /* Wide comment */
 #define LINE_COMMENT 2         /* One-line comment */

 #define MIN(a,b) ((a < b)? a: b)

 #define PUSH_NODE(node) \
        stack[sp] = node; \
        sp++;

 #define POP_NODE \
        sp--;

 #define TOS_NODE \
        stack[sp]

 #define PUSH_LEVEL(level) \
        lstack[lsp] = level; \
        lsp++;

 #define POP_LEVEL \
        lsp--;

 #define TOS_LEVEL \
        lstack[lsp]

 /* Common variables */
 xfChar s; xfInt i;
 /* Last character, comment and long string flags */
 xfInt last = 0, comment = false, string = false;
 /* precomments shows if first comment parenthesis characted
    was found */
 xfInt precomment = false;
 /* Source byte count, line and column */
 xfInt pos = 0,line = 1,col = 1;
 /* String buffer */
 xfChar *word;
 /* Node and node level stacks and their sizes */
 xfInt sp = 0, lsp = 0;
 xfNode **stack; xfInt *lstack;
 /* String type flags: node, attribute, class */
 xfInt justnode = false, isattr = false, isclass = false;
 /* True if current node class/value was defined */
 xfInt defclass = false,defvalue = false;
 /* Last added node pointer */
 xfNode *lastnode = XF_NULL;
 /* Current node parent */
 xfNode *parent = XF_NULL,*cparent;
 /* Current node incapsulation level */
 xfInt level = 0;

 if (map) cparent = map->root;

 /* Allocate string buffer */
 word = (xfChar*)xfAlloc(XF_NODESIZE * sizeof(xfChar));
 if (!word) return XFE_MEMORY;

 /* Allocate node stack */
 stack = (xfNode**)xfAlloc(XF_STACKSIZE * sizeof(xfNode*));
 if (!stack) {
   xfFree(word);
   return XFE_MEMORY;
  }

 /* Allocate level stack (for event-based parsing) */
 lstack = (xfInt*)xfAlloc(XF_STACKSIZE * sizeof(xfInt*));
 if (!lstack) {
   xfFree(word);
   xfFree(stack);
   return XFE_MEMORY;
  }

 /*
   Set error line, column and message to zeroes
 */
 if (map) {
   xfSetErrorLine(map,0);
   xfSetErrorCol(map,0);
   xfSetError(map,XF_NULL);
  }

 /*
   Check for document size limit (security system)
 */
 if (map) {
   if (map->security) {
     i = map->security->docSize;
     if (i)
       if (xfReaderGetSize(reader) > i)
          ERROR("SecurityLimit");
    }
  }

 /*
   Yeah... parse this document!
 */
 do
  {
   /* Clear string buffer */
   memset(word, 0, XF_NODESIZE * sizeof(xfChar));

   /* If we have last character in one-byte buffer, use it */
   if (last)
    { word[0] = last; last = 0; }
   else
    do
     {
      /* Fast reading of UTF-16 character from memory */
      if (reader->type == XF_READER_MEMORY)
        {
         s = reader->string[reader->pos];
	 if (!s) break;
         reader->pos++;
        }
      else s = xfReaderInputChar(reader);

      /* Hmmm... that is the end of file/string */
      if (s == '\x00')
       {
        if (comment == WIDE_COMMENT) ERROR("UnterminatedComment");
        if (string) ERROR("UnterminatedString");
        if (sp) ERROR("UnterminatedBlock");
        break;
       }

      /* Increase line/column value */
      col++;
      if (s == '\x0A') { line++; col = 0; }

      /* Ignore everithing inside one-line comment */
      if (comment == LINE_COMMENT)
        { if (s == L'\x0A' || s == L'\x0D') comment = false; continue; }

      /* Ignore everithing inside long comment */
      if (comment == WIDE_COMMENT)
       {
        if (!precomment)
         {
          if (s == '*') precomment = true;
         }
        else
         {
          if (s == '/') { precomment = false; comment = false; continue; }
          if (s != '*') precomment = false;
         }
        continue;
       }

      /* If we have long string, unpack it */
      if (string)
       {
        if (s == '\"')
         {
          string = false;
          xfUnpackString(word + pos);
          continue;
         }
        else
         {
          if (xfStrlen(word) + 1 < XF_NODESIZE)
           word[xfStrlen(word)] = s;
          }
        continue;
       }

      /* Some special character which ends string */
      if (xfStrchr(XFS(" \t\x0A\x0D"),s)) {
        if (*word) break;
        continue;
       }

      /* Process begin/end of string */
      if (s == '\"')
       {
        string = true; if (*word) pos = xfStrlen(word); else pos = 1;
        if (!*word) word[0] = '\"';
        continue;
       }

      /* Send special character to a single-byte buffer and process
         string buffer */
      if (xfStrchr(XFS("{}\n*[/;=:"), s))
       { if (*word) last = s; else { word[0] = s; } break; }

      /* Check if character is correct */
      if (!isalnum(s) && !xfStrchr(XFS("\"_-"),s))
       ERROR("IncorrectChar");

      if (!*word) { word[0] = '\"'; }
      if (xfStrlen(word) + 1 < XF_NODESIZE)
       {
        i = xfStrlen(word);
        word[i] = s;
        word[i + 1] = (xfChar)0x0000;
       }
     }
   while (1);

   if (!*word && s) continue;

   if (word[0] == '/')
    { precomment = true;
      continue; }

   if (precomment)
    {
     if (word[0] == '*')
      {
       comment = WIDE_COMMENT;
       precomment = false;
       continue;
      }
     if (word[0] != '/') precomment = false;
    }

   if (word[0] == '*') { comment = LINE_COMMENT; continue; }

   ////////////////////////////////////////////////////////////////////////////

   if (!*word) continue;

   /* New level of incapsulation (increment stack) */
   if (word[0] == '{')
    {
     if (sp >= XF_STACKSIZE) ERROR("StackOverflow");

     PUSH_NODE(parent);

     parent = cparent;
     justnode = false;
     continue;
    }

   /* Close one incapsulation level (decrement stack) */
   if (word[0] == '}')
    {
     if (!sp) ERROR("UnexpectedCloseBlock");

     POP_NODE;
     parent = TOS_NODE; cparent = parent;

     POP_LEVEL;
     level = TOS_LEVEL;
     
     justnode = false;
     continue;
    }

   /* End of sentencse */
   if (word[0] == ';')
    {
     cparent = parent;

     POP_LEVEL;
     level = TOS_LEVEL;

     justnode = false;
     continue;
    }

   /* Begin attribute */
   if (word[0] == '=')
    {
     if (!justnode || isattr || isclass || defvalue) ERROR("UnexpectedValueDef");
     isattr = true;
     continue;
    }

   if (word[0] == ':')
    {
     if (!justnode || isattr || isclass || defclass) ERROR("UnexpectedClassDef");
     isclass = true;
     continue;
    }

   /* Node */

   if (xfStrlen(word) < 2)
    {
     if (isattr) ERROR("EmptyNodeValue");
     if (isclass) ERROR("EmptyNodeClass");
     ERROR("EmptyNodeName");
    }

   if (!isattr && !isclass)
    {
     /* Add node to the map */
     if (map) {
       if (!xfCanAddNode(map)) ERROR("SecurityLimit");

       if (cparent)
         lastnode = xfAdd(cparent, word + 1,XF_NULL,XF_NULL);
       else
         lastnode = xfAdd(map->root,word + 1,XF_NULL,XF_NULL);
        }

     /* Open node event */
     level++;
     PUSH_LEVEL(level);

     if (map) {
       if (!lastnode) ERROR("NotEnoughMemory");
      }

     justnode = true;
     defvalue = defclass = false;
     cparent = lastnode;
    }
   else
    {
     /* Node value */
     if (isattr)
      {
       if (map) xfSet(lastnode, word + 1);
       isattr = false;
       isclass = false;
       defvalue = true;
      }

     /* Node class */
     if (isclass)
      {
       if (map) xfSetClass(lastnode, word + 1);
       isattr = false;
       isclass = false;
       defclass = true;
      }
    }
  }
 while (s);

 xfFree(word);
 xfFree(stack);
 xfFree(lstack);

 return XFE_OK;
 #undef ERROR(msg)
}


