/*
  SimTest - Simple Benchmark Library
  (C) Max Volkov (xfsoft@mail.ru), Sevastopol 2006
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "simtest.h"

int st_error = 0;
char *st_name = 0;
st_float st_tick = 0;
int st_ops = 0;

/* Default number of tests and repeats per test
   If one of those variables is not set by a program,
   we use default SimTest values */
int st_tests = 0;
int st_repeats = 0;

void st_init()
{
 int a = CLK_TCK;
 st_tick = (st_float)a;
 if (st_tick <= 0) st_crash("incorrect CLK_TCK value");
}

st_float st_time()
{
 return (st_float)clock();
}

void st_crash(char *msg)
{
 printf("Error: %s\n",msg);
 exit(1);
}

int st_work(int tests,int repeats)
{
 int test,i,error;
 st_float time1,time2,dt;
 st_float mintime,maxtime,avgtime,alltime;
 char *msg[100];

 mintime = maxtime = alltime = 0e1;

 printf("Benchmark info:\n");
 printf("   Number of tests: %i\n",tests);
 printf("   Repeats per test: %i\n",repeats);
 printf("   Operations per st_test() call: %i\n",st_ops);
 printf("   Clock resolution: %1.4f ticks per sec\n\n",st_tick);

 printf("Testing: please wait...\n\n");

 for (test = 0; test < tests; test++)
  {
   time1 = st_time();

   for (i = 0; i < repeats; i++) st_test();

   time2 = st_time();

   if (st_error)
    {
     sprintf(msg,"test failled (function returned error code %i)",st_error);
     st_crash(msg);
    }

   dt = time2 - time1;
   if (time2 == time1) st_crash("too short period of time");

   alltime += dt;

   if (test)
    {
     if (dt < mintime) mintime = dt;
     if (dt > maxtime) maxtime = dt;
    }
   else mintime = maxtime = dt;
  }

 avgtime = alltime / tests;

 st_log(avgtime,mintime,maxtime,tests,repeats,st_ops);

 return 0;
}

void st_log(st_float avgtime,st_float mintime,st_float maxtime,int tests,
            int repeats,int ops)
{
 #define ST_ALIGN(a,c) \
  if (a > 1000) { a /= 1000; c = 'K'; } \
  if (a > 1000) { a /= 1000; c = 'M'; } \
  if (a > 1000) { a /= 1000; c = 'G'; }

 st_float avgperf,minperf,maxperf;
 st_float _avgperf,_minperf,_maxperf;
 char avg_c,min_c,max_c;
 avg_c = min_c = max_c = ' ';

 /*
   Log file format
   Log lines start with #, other infornation is human-readable

   % test name
   # avg-performance min-performance max-performance num-tests num-repeats \
     ops_per-test clock-resolution
 */

 if (!st_tick) st_crash("undefined st_tick value");
 if (!st_ops) st_crash("undefined st_ops value");

 avgperf = ((1 / avgtime) * st_tick) * st_ops * repeats;
 minperf = ((1 / maxtime) * st_tick) * st_ops * repeats;
 maxperf = ((1 / mintime) * st_tick) * st_ops * repeats;

 _avgperf = avgperf;
 _minperf = minperf;
 _maxperf = maxperf;

 ST_ALIGN(avgperf,avg_c);
 ST_ALIGN(minperf,min_c);
 ST_ALIGN(maxperf,max_c);

 /* Human-readable info */
 printf("Performance:\n");
 printf("   Average:\t%1.4f %cops\n",avgperf,avg_c);
 printf("   Minimal:\t%1.4f %cops\n",minperf,min_c);
 printf("   Maximal:\t%1.4f %cops\n\n",maxperf,max_c);

 /* Machine-readable info */
 if (st_name) printf("%% %s\n",st_name); else printf("%%\n");
 printf("# %f %f %f %i %i %i %f\n",_avgperf,_minperf,_maxperf,tests,repeats,
        st_ops,st_tick);
}

int main(int argc,char *argv[])
{
 printf("SimTest %i.%i\n",ST_VN,ST_SN);
 printf("Usage:\n");
 printf("   ./test\n");
 printf("   ./test num-tests\n");
 printf("   ./test num-tests num-repeats\n\n");

 if (st_tests <= 0) st_tests = ST_TESTS;
 if (st_repeats <= 0) st_repeats = ST_REPEATS;

 st_init();

 st_start();

 if (argc == 2) st_repeats = atol(argv[1]);
 if (argc == 3) {
  st_tests = atol(argv[1]);
  st_repeats = atol(argv[2]); }

 st_work(st_tests,st_repeats);
 st_stop();

 return 0;
}

